/* SPIM S20 MIPS simulator.
   Code to create, maintain and access memory.
   Copyright (C) 1990 by James Larus (larus@cs.wisc.edu).

   SPIM is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 1, or (at your option) any
   later version.

   SPIM is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with GNU CC; see the file COPYING.  If not, write to James R.
   Larus, Computer Sciences Department, University of Wisconsin--Madison,
   1210 West Dayton Street, Madison, WI 53706, USA or to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */


/* $Header: /var/home/larus/SPIM/RCS/mem.c,v 3.10 1991/05/09 21:00:44 larus Exp larus $
*/


#include "spim.h"
#include "inst.h"
#include "mem.h"
#include "reg.h"


/* Exported functions: */

void bad_mem_read ();
void bad_mem_write ();
void bad_text_read ();
void bad_text_write ();
void expand_data ();
void expand_k_data ();
void expand_stack ();
void expand_text ();
void make_memory ();
void print_mem ();


/* Imported functions: */

void print_inst ();
char *realloc ();


/* Local functions: */

static void free_instructions ();



/* Memory is allocated in five chunks:
	text, data, stack, kernel text, and kernel data.

   The arrays are independent and have different semantics.

   text is allocated from 0x400000 up and only contains INSTRUCTIONs.

   data is allocated from 0x10000000 up.  It can be extended by the
   SBRK system call.  Programs can only read and write this segment.

   stack grows from 0x7fffefff down.  It is automatically extended.
   Programs can only read and write this segment.

   kernel text is allocated from 0x80000000 up and only contains
   INSTRUCTIONs.

   k_data is like data, but is allocated from 0x90000000 up.

   Both kernel text and kernel data can only be accessed in kernel mode.
*/


void
make_memory (text_size, data_size, stack_size, k_text_size, k_data_size)
     long text_size, data_size, stack_size, k_text_size, k_data_size;
{
  if (data_size <= 65536)
    fatal_error ("Data segment must be larger than 64K\n");
  if (text_seg == NULL)
    text_seg = (instruction **)
      malloc (sizeof (instruction *) * text_size / BYTES_PER_WORD);
  else
    free_instructions (text_seg, (text_top - TEXT_BOT) / BYTES_PER_WORD);
  bzero (text_seg, sizeof (instruction *) * text_size / BYTES_PER_WORD);
  text_top = TEXT_BOT + text_size;

  if (data_seg == NULL)
    data_seg =
      (mem_word *) malloc (sizeof (mem_word) * data_size / BYTES_PER_WORD);
  bzero (data_seg, sizeof (instruction *) * data_size / BYTES_PER_WORD);
  data_seg_b = (BYTE_TYPE *) data_seg;
  data_seg_h = (short *) data_seg;
  data_top = DATA_BOT + data_size;

  if (stack_seg == NULL)
    stack_seg =
      (mem_word *) malloc (sizeof (mem_word) * stack_size / BYTES_PER_WORD);
  bzero (stack_seg, sizeof (instruction *) * stack_size / BYTES_PER_WORD);
  stack_seg_b = (BYTE_TYPE *) stack_seg;
  stack_seg_h = (short *) stack_seg;
  stack_bot = STACK_TOP - stack_size;

  if (k_text_seg == NULL)
    k_text_seg = (instruction **)
      malloc (sizeof (instruction *) * k_text_size / BYTES_PER_WORD);
  else
    free_instructions (k_text_seg, (k_text_top - K_TEXT_BOT) / BYTES_PER_WORD);
  bzero (k_text_seg, sizeof (instruction *) * k_text_size / BYTES_PER_WORD);
  k_text_top = K_TEXT_BOT + k_text_size;

  if (k_data_seg == NULL)
    k_data_seg =
      (mem_word *) malloc (sizeof (mem_word) * k_data_size / BYTES_PER_WORD);
  bzero (k_data_seg, sizeof (instruction *) * k_data_size / BYTES_PER_WORD);
  k_data_seg_b = (BYTE_TYPE *) k_data_seg;
  k_data_seg_h = (short *) k_data_seg;
  k_data_top = K_DATA_BOT + k_data_size;

  text_modified = 0;
  data_modified = 0;
}


/* Free the storage used by the old instructions in memory. */

static void
free_instructions (inst, n)
     register instruction **inst;
     int n;
{
  for ( ; n > 0; n --, inst ++)
    if (*inst)
      {
	if (EXPR (*inst))
	  {
	    /*if (EXPR (*inst)->symbol)
	      free (EXPR (*inst)->symbol);*/
	    free (EXPR (*inst));
	  }
	free (*inst);
      }
}


/* Expand the text segment by adding N bytes. */

void
expand_text (addl_bytes, kflag)
     long addl_bytes;
     int kflag;
{
  register instruction **p;
  instruction ***s;
  int n;

  if (source_file && (addl_bytes < 0 || addl_bytes > MAX_SEG_GROWTH))
    run_error ("Memory access %d bytes beyond text segment (cannot expand that far)\n",
	       addl_bytes);
  if (kflag)
    {
      s = &k_text_seg;
      n = k_text_top - K_TEXT_BOT;
      k_text_top += addl_bytes;
    }
  else
    {
      s = &text_seg;
      n = text_top - TEXT_BOT;
      text_top += addl_bytes;
    }
  *s = (instruction **) realloc (*s, n + addl_bytes);
  for (p = (*s) + n / BYTES_PER_WORD;
       addl_bytes > 0;
       addl_bytes -= BYTES_PER_WORD)
    *p ++ = NULL;
}


/* Expand the data segment by adding N bytes. */

void
expand_data (addl_bytes)
     long addl_bytes;
{
  register mem_word *p;

  if (source_file && (addl_bytes < 0 || addl_bytes > MAX_SEG_GROWTH))
    run_error ("Memory access %d bytes beyond data segment (cannot expand that far)\n",
	       addl_bytes);
  data_seg = (mem_word *) realloc (data_seg, data_top - DATA_BOT + addl_bytes);
  data_seg_b = (BYTE_TYPE *) data_seg;
  data_seg_h = (short *) data_seg;
  for (p = data_seg + (data_top - DATA_BOT) / BYTES_PER_WORD;
       p < data_seg + (data_top - DATA_BOT + addl_bytes) / BYTES_PER_WORD; )
    *p ++ = 0;
  data_top += addl_bytes;
}


/* Expand the stack segment by adding N bytes.
   Can't use REALLOC since it copies from bottom of memory blocks and
   stack grows down from top of its block. */

void
expand_stack (addl_bytes)
     long addl_bytes;
{
  long old_size = STACK_TOP - stack_bot;

  mem_word *new_seg = (mem_word *) malloc (old_size + addl_bytes);
  register mem_word *po = stack_seg + old_size / BYTES_PER_WORD;
  register mem_word *pn = new_seg + (old_size + addl_bytes) / BYTES_PER_WORD;

  if (source_file && (addl_bytes < 0 || addl_bytes > MAX_SEG_GROWTH))
    run_error ("Memory access %d bytes below stack segment (cannot expand that far)\n",
	       addl_bytes);
  for ( ; po >= stack_seg ; ) *pn -- = * po --;
  for ( ; pn >= new_seg ; ) *pn -- = 0;

  stack_seg = new_seg;
  stack_seg_b = (BYTE_TYPE *) stack_seg;
  stack_seg_h = (short *) stack_seg;
  stack_bot -= addl_bytes;
}


/* Expand the kernel data segment by adding N bytes. */

void
expand_k_data (addl_bytes)
     long addl_bytes;
{
  register mem_word *p;

  if (source_file && (addl_bytes < 0 || addl_bytes > MAX_SEG_GROWTH))
    run_error ("Memory access %d bytes beyond kernel data segment (cannot expand that far)\n",
	       addl_bytes);
  k_data_seg = (mem_word *) realloc (k_data_seg,
				     k_data_top - K_DATA_BOT + addl_bytes);
  k_data_seg_b = (BYTE_TYPE *) k_data_seg;
  k_data_seg_h = (short *) k_data_seg;
  for (p = k_data_seg + (k_data_top - K_DATA_BOT) / BYTES_PER_WORD;
       p < k_data_seg + (k_data_top - K_DATA_BOT + addl_bytes) /BYTES_PER_WORD;
       )
    *p ++ = 0;
  k_data_top += addl_bytes;
}


void
bad_text_read (addr)
     mem_addr addr;
{
  RAISE_EXCEPTION ((addr & 0x3) ? ADDRL_EXCPT : IBUS_EXCPT, BadVAddr = addr);
}


void
bad_text_write (addr, inst)
     mem_addr addr;
     instruction *inst;
{
  if (addr & 0x3)
    RAISE_EXCEPTION (ADDRL_EXCPT, BadVAddr = addr)
  else if (addr < TEXT_BOT)
    RAISE_EXCEPTION (IBUS_EXCPT, BadVAddr = addr)
  else
    {
      expand_text (TEXT_CHUNK_SIZE, addr >= K_TEXT_BOT);
      if (addr >= K_TEXT_BOT)
	{
	  if (addr <= k_text_top)
	    k_text_seg [(addr - K_TEXT_BOT) >> 2] = inst;
	  else
	    RAISE_EXCEPTION (IBUS_EXCPT, BadVAddr = addr)
	}
      else
	{
	  if (addr <= text_top)
	    text_seg [(addr - TEXT_BOT) >> 2] = inst;
	  else
	    RAISE_EXCEPTION (IBUS_EXCPT, BadVAddr = addr)
	}
    }
}


void
bad_mem_read (addr, mask, dest)
     mem_addr addr;
     int mask;
     mem_word *dest;
{
  if (addr & mask)
    RAISE_EXCEPTION (ADDRL_EXCPT, BadVAddr = addr)
  else if (addr < TEXT_BOT)
    RAISE_EXCEPTION (IBUS_EXCPT, BadVAddr = addr)
  else if (addr < DATA_BOT || addr >= STACK_TOP)
    /* Address out of range: below data or above stack */
    RAISE_EXCEPTION (DBUS_EXCPT, BadVAddr = addr)
  else if (addr - data_top < stack_bot - addr)
    {
      /* Address closer to data segment, so grow data segment */
      expand_data (addr == data_top ? MAX_SEG_GROWTH : (addr - data_top) * 2);
      *dest = 0;		/* Newly allocated memory */
    }
  else
    {
      /* Grow stack segment */
      expand_stack (addr == stack_bot ? MAX_SEG_GROWTH
		    : (stack_bot - addr) * 2);
      *dest = 0;		/* Newly allocated memory */
    }
}


void
bad_mem_write (addr, value, mask)
     mem_addr addr;
     mem_word value;
     int mask;
{
  if (addr & mask)
    /* Unaligned address fault */
    RAISE_EXCEPTION (ADDRS_EXCPT, BadVAddr = addr)
  else if (addr < TEXT_BOT)
    RAISE_EXCEPTION (IBUS_EXCPT, BadVAddr = addr)
  else if (addr < DATA_BOT || addr >= STACK_TOP)
    /* Address out of range: below data or above stack */
    RAISE_EXCEPTION (DBUS_EXCPT, BadVAddr = addr)
  /* Address between data and stack */
  else if (addr - data_top <  stack_bot - addr)
    {
      /* Address closer to data segment, so grow data segment */
      expand_data (addr == data_top ? MAX_SEG_GROWTH : (addr - data_top) * 2);
      if (addr <= data_top)
	{
	  if (mask == 0)
	    data_seg_b [addr - DATA_BOT] = value;
	  else if (mask == 1)
	    data_seg_h [addr - DATA_BOT] = value;
	  else
	    data_seg [addr - DATA_BOT] = value;
	}
      else
	RAISE_EXCEPTION (DBUS_EXCPT, BadVAddr = addr)
    }
  else
    {
      /* Grow stack segment */
      expand_stack (addr == stack_bot ? MAX_SEG_GROWTH
		    : (stack_bot - addr) * 2);
      if (addr >= stack_bot)
	{
	  if (mask == 0)
	    stack_seg_b [addr - stack_bot] = value;
	  else if (mask == 1)
	    stack_seg_h [addr - stack_bot] = value;
	  else
	    stack_seg [addr - stack_bot] = value;
	}
      else
	RAISE_EXCEPTION (DBUS_EXCPT, BadVAddr = addr)
    }
}


void
print_mem (addr)
     mem_addr addr;
{
  long value;

  if (TEXT_BOT <= addr && addr < text_top)
    print_inst (addr);
  else if (DATA_BOT <= addr && addr < data_top)
    {
      READ_MEM_WORD (value, addr);
      write_output (message_out, "Data seg @ 0x%08x (%d) = 0x%08x (%d)\n",
		    addr, addr, value, value);
    }
  else if (stack_bot <= addr && addr < STACK_TOP)
    {
      READ_MEM_WORD (value, addr);
      write_output (message_out, "Stack seg @ 0x%08x (%d) = 0x%08x (%d)\n",
		    addr, addr, value, value);
    }
  else if (K_TEXT_BOT <= addr && addr < k_text_top)
    print_inst (addr);
  else if (K_DATA_BOT <= addr && addr < k_data_top)
    {
      READ_MEM_WORD (value, addr);
      write_output (message_out,
		    "Kernel Data seg @ 0x%08x (%d) = 0x%08x (%d)\n",
		    addr, addr, value, value);
    }
  else
    error ("Address to print_mem is out of bounds: 0x%08x (%d)\n", addr, addr);
}
