/* SPIM S20 MIPS simulator.
   Terminal interface for SPIM simulator.
   Copyright (C) 1990 by James Larus (larus@cs.wisc.edu).

   SPIM is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 1, or (at your option) any
   later version.

   SPIM is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License
   along with GNU CC; see the file COPYING.  If not, write to James R.
   Larus, Computer Sciences Department, University of Wisconsin--Madison,
   1210 West Dayton Street, Madison, WI 53706, USA or to the Free
   Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */


/* $Header: /var/home/larus/SPIM/RCS/spim.c,v 3.9 1991/05/05 19:16:11 larus Exp larus $
*/


#include <stdio.h>
#include <ctype.h>
#include <varargs.h>
#include <setjmp.h>
#include <signal.h>

#include "spim.h"
#include "inst.h"
#include "mem.h"
#include "reg.h"
#include "y.tab.h"


/* Exported functions: */

void error ();
void read_input ();
int run_error ();
void write_output ();


/* Imported functions: */

void add_breakpoint ();
void delete_breakpoint ();
void exit ();
mem_addr find_symbol_address ();
void initialize_scanner ();
void initialize_world ();
void list_breakpoints ();
void print_mem ();
void print_symbols ();
int read_assembly_file ();
int read_aout_file ();
int run_program ();
mem_addr starting_addr ();
int vfprintf ();
int yylex ();
int yyparse ();


/* Internal functions: */

static void control_c_seen ();
static void flush_to_newline ();
static long get_opt_int ();
static int parse_spim_command ();
static int print_reg ();
static int read_assembly_command ();
static int str_prefix ();
static void top_level ();


/* Imported variables: */

mem_addr program_starting_address; /* First PC */
int yylval;			/* Value of token from YYLEX */
char *input_file_name;


/* Local variables: */

static int load_trap_handler = 1; /* Non-zero means load standard trap
				     handler */

static jmp_buf spim_top_level_env; /* For ^C */



int
main (argc, argv)
     int argc;
     char **argv;
{
  int i;
  char *file_name = NULL, *ex_file_name = NULL;

  console_out = (int) stdout;
  message_out = (int) stdout;

  for (i = 1; i < argc; i++)
    if (streq (argv [i], "-bare"))
      bare_machine = 1, quiet = 1;
    else if (streq (argv [i], "-asm"))
      bare_machine = 0;
    else if (streq (argv [i], "-trap"))
      load_trap_handler = 1;
    else if (streq (argv [i], "-notrap"))
      load_trap_handler = 0;
    else if (streq (argv [i], "-quiet"))
      quiet = 1;
    else if (streq (argv [i], "-noquiet"))
      quiet = 0;
    else if (streq (argv [i], "-file"))
      file_name = argv[++i];
#ifdef mips
    else if (streq (argv [i], "-execute"))
      ex_file_name = argv[++i];
#endif
    else
      error ("usage: spim -bare/-asm -trap/-notrap -quiet/-noquiet -file/-execute <file>\n");

  initialize_world (load_trap_handler);
  if ((file_name == NULL) && (ex_file_name == NULL))
    top_level ();
  else
    {
#ifdef mips
      if (ex_file_name)
	{
	  read_aout_file (ex_file_name);
	  run_program (program_starting_address, DEFAULT_RUN_STEPS, 0, 0);
	}
      else
#endif
	{
	  if (!read_assembly_file (file_name))
	    run_program (find_symbol_address (DEFAULT_RUN_LOCATION),
			 DEFAULT_RUN_STEPS, 0, 0);
	}
    }
  return (0);
}


/* Top-level read-eval-print loop for SPIM. */

static void
top_level ()
{
  int redo = 0;			/* Non-zero means reexecute last command */

  signal (SIGINT, control_c_seen);
  while (1)
    {
      if (!redo)
	write_output (message_out, "(spim) ");
      if (!setjmp (spim_top_level_env))
	redo = parse_spim_command (stdin, redo);
      else
	redo = 0;
      fflush (stdout);
      fflush (stderr);
    }
}


static void
control_c_seen ()
{
  write_output (message_out, "Execution interrupted\n");
  longjmp (spim_top_level_env, 1);
}


/* SPIM commands */

#define UNKNOWN_CMD		0
#define EXIT_CMD		1
#define READ_CMD		2
#define RUN_CMD			3
#define STEP_CMD		4
#define PRINT_CMD		5
#define PRINT_SYM_CMD		6
#define REINITIALIZE_CMD	7
#define ASM_CMD			8
#define REDO_CMD		9
#define NOP_CMD			10
#define HELP_CMD		11
#define CONTINUE_CMD		12
#define SET_BKPT_CMD		13
#define DELETE_BKPT_CMD		14
#define LIST_BKPT_CMD		15
#ifdef mips
#define EXECUTE_CMD		16
#endif

/* Parse a SPIM command from the FILE and execute it.  If REDO is non-zero,
   don't read a new command; just rexecute the previous one.
   Return non-zero if the command was to redo the previous command. */

static int
parse_spim_command (file, redo)
     FILE *file;
     int redo;
{
  static int prev_cmd = NOP_CMD; /* Default redo */
  static int prev_token;
  int cmd;

  initialize_scanner (file);
  switch (cmd = (redo ? prev_cmd : read_assembly_command (file)))
    {
    case EXIT_CMD:
      exit (0);

    case READ_CMD:
      {
	int token = (redo ? prev_token : yylex ());

	if (!redo) flush_to_newline ();
	if (token == Y_STR)
	  {
	    read_assembly_file ((char *) yylval);
	    initialize_scanner (file); /* Reinitialize! */
	  }
	else
	  error ("Must supply a filename to read\n");
	prev_cmd = READ_CMD;
	return (0);
      }

    case RUN_CMD:
      {
	static mem_addr addr;

	addr = (redo ? addr : get_opt_int ());
	if (addr == 0)
	  addr = starting_address ();

	if (addr)
	  if (run_program (addr, DEFAULT_RUN_STEPS, 0, 0))
	    write_output (message_out, "Breakpoint encountered at 0x%08x\n",
			  PC);

	prev_cmd = RUN_CMD;
	return (0);
      }

    case CONTINUE_CMD:
      {
	if (PC != 0)
	  if (run_program (PC, DEFAULT_RUN_STEPS, 0, 1))
	    write_output (message_out, "Breakpoint encountered at 0x%08x\n",
			  PC);

	prev_cmd = CONTINUE_CMD;
	return (0);
      }

    case STEP_CMD:
      {
	static int steps;
	mem_addr addr;

	steps = (redo ? steps : get_opt_int ());
	addr = starting_address ();

	if (steps == 0)
	  steps = 1;
	if (addr)
	  if (run_program (addr, steps, 1, 1))
	    write_output (message_out, "Breakpoint encountered at 0x%08x\n",
			  PC);

	prev_cmd = STEP_CMD;
	return (0);
      }

    case PRINT_CMD:
      {
	int token = (redo ? prev_token : yylex ());
	static int loc;

	if (token == Y_REG)
	  {
	    if (redo) loc += 1;
	    else loc = yylval;
	    print_reg (loc, 0);
	  }
	else if (token == Y_FP_REG)
	  {
	    if (redo) loc += 2;
	    else loc = yylval;
	    print_reg (loc, 1);
	  }
	else if (token == Y_INT)
	  {
	    if (redo) loc += 4;
	    else loc = yylval;
	    print_mem (loc);
	  }
	else if (token == Y_ID)
	  {
	    if (!print_reg (yylval, 2))
	      {
		if (redo) loc += 4;
		else loc = find_symbol_address ((char *) yylval);

		if (loc != 0)
		  print_mem (loc);
		else
		  error ("Unknown label: %s\n", yylval);
	      }
	  }
	else
	  error ("Print what?\n");
	if (!redo) flush_to_newline ();
	prev_cmd = PRINT_CMD;
	prev_token = token;
	return (0);
      }

    case PRINT_SYM_CMD:
      print_symbols ();
      if (!redo) flush_to_newline ();
      prev_cmd = NOP_CMD;
      return (0);

    case REINITIALIZE_CMD:
      initialize_world (load_trap_handler && !bare_machine);
      prev_cmd = NOP_CMD;
      return (0);

    case ASM_CMD:
      input_file_name = "<standard input>";
      yyparse ();
      prev_cmd = ASM_CMD;
      return (0);

    case REDO_CMD:
      return (1);

    case NOP_CMD:
      prev_cmd = NOP_CMD;
      return (0);

    case HELP_CMD:
      if (!redo) flush_to_newline ();
      write_output (message_out, "\nSPIM is a MIPS R2000 simulator.\n");
      write_output (message_out, "Its top-level commands are:\n");
      write_output (message_out, "exit  -- Exit from the simulator\n");
      write_output (message_out,
		 "read \"FILE\" -- Read FILE of assembly code into memory\n");
      write_output (message_out,
		 "load \"FILE\" -- Same as read\n");
#ifdef mips
      write_output (message_out,
		 "execute \"A.OUT\" -- Read A.OUT file into memory \n");
#endif
      write_output (message_out,
		 "run <ADDR> -- Start the program at optional ADDRESS\n");
      write_output (message_out,
		 "step <N> -- Step the program for N instructions\n");
      write_output (message_out,
		 "continue -- Continue program execution without stepping\n");
      write_output (message_out, "print $N -- Print register N\n");
      write_output (message_out,
		 "print $fN -- Print floating point register N\n");
      write_output (message_out,
		 "print ADDR -- Print contents of memory at ADDRESS\n");
      write_output (message_out,
		 "reinitialize -- Clear the memory and registers\n");
      write_output (message_out,
		 "breakpoint <ADDR> -- Set a breakpoint at address\n");
      write_output (message_out,
		 "delete <ADDR> -- Delete all breakpoints at address\n");
      write_output (message_out, "list -- List all breakpoints\n");
      write_output (message_out,
		 ". -- Rest of line is assembly instruction to put in memory\n");
      write_output (message_out, "<cr> -- Newline reexecutes previous command\n");
      write_output (message_out, "? -- Print this message\n");
      write_output (message_out,
		 "\nMost commands can be abbreviated to their unique prefix\n");
      write_output (message_out, "e.g., ex, re, l, ru, s, p\n\n");
      prev_cmd = HELP_CMD;
      return (0);

    case SET_BKPT_CMD:
    case DELETE_BKPT_CMD:
      {
	int token = (redo ? prev_token : yylex ());
	static mem_addr addr;

	if (!redo) flush_to_newline ();
	if (token == Y_INT)
	  addr = (redo ? addr + 4 : yylval);
	else
	  error ("Must supply an address for breakpoint\n");
	if (cmd == SET_BKPT_CMD)
	  add_breakpoint (addr);
	else
	  delete_breakpoint (addr);
	prev_cmd = cmd;
	return (0);
      }

    case LIST_BKPT_CMD:
      if (!redo) flush_to_newline ();
      list_breakpoints ();
      prev_cmd = LIST_BKPT_CMD;
      return (0);

#ifdef mips
    case EXECUTE_CMD:
      {
	int token = (redo ? prev_token : yylex ());

	if (!redo) flush_to_newline ();
	if (token == Y_STR)
	  read_aout_file ((char *) yylval);
	else
	  error ("Must supply a filename to read\n");
	prev_cmd = EXECUTE_CMD;
	return (0);
      }
#endif

    default:
      while (yylex () != Y_NL) ;
      error ("Unknown spim command\n");
      return (0);
    }
}


/* Read a SPIM command from the FILE and return its ennuemerated value. */

static int
read_assembly_command (file)
     FILE *file;
{
  int token = yylex ();

  if (token == Y_NL)		/* Blank line means redo */
    return (REDO_CMD);
  else if (token != Y_ID)	/* Better be a string */
    return (UNKNOWN_CMD);
  else if (str_prefix ((char *) yylval, "exit", 2))
    return (EXIT_CMD);
  else if (str_prefix ((char *) yylval, "print", 1))
    return (PRINT_CMD);
  else if (str_prefix ((char *) yylval, "print_symbol", 6))
    return (PRINT_SYM_CMD);
  else if (str_prefix ((char *) yylval, "run", 2))
    return (RUN_CMD);
  else if (str_prefix ((char *) yylval, "read", 2))
    return (READ_CMD);
  else if (str_prefix ((char *) yylval, "load", 2))
    return (READ_CMD);
  else if (str_prefix ((char *) yylval, "reinitialize", 6))
    return (REINITIALIZE_CMD);
  else if (str_prefix ((char *) yylval, "step", 1))
    return (STEP_CMD);
  else if (str_prefix ((char *) yylval, "help", 1))
    return (HELP_CMD);
  else if (str_prefix ((char *) yylval, "continue", 1))
    return (CONTINUE_CMD);
  else if (str_prefix ((char *) yylval, "breakpoint", 2))
    return (SET_BKPT_CMD);
  else if (str_prefix ((char *) yylval, "delete", 1))
    return (DELETE_BKPT_CMD);
  else if (str_prefix ((char *) yylval, "list", 2))
    return (LIST_BKPT_CMD);
#ifdef mips
  else if (str_prefix ((char *) yylval, "execute", 3))
    return (EXECUTE_CMD);
#endif
  else if (*(char *) yylval == '?')
    return (HELP_CMD);
  else if (*(char *) yylval == '.')
    return (ASM_CMD);
  else
    return (UNKNOWN_CMD);
}


/* Return non-nil if STRING1 is a (proper) prefix of STRING2. */

static int
str_prefix (s1, s2, min_match)
     char *s1, *s2;
     int min_match;
{
  for ( ; *s1 == *s2 && *s1 != '\0'; s1 ++, s2 ++) min_match --;
  return (*s1 == '\0' && min_match <= 0);
}


/* Read and return an integer from the current line of input.  If the
   line doesn't contain an integer, return 0.  In either case, flush the
   rest of the line, including the newline. */

static long
get_opt_int ()
{
  int token;

  if ((token = yylex ()) == Y_INT)
    {
      flush_to_newline ();
      return (yylval);
    }
  else if (token == Y_NL)
    return (0);
  else
    {
      flush_to_newline ();
      return (0);
    }
}


/* Flush the rest of the input line up to and including the next newline. */

static void
flush_to_newline ()
{
  while (yylex () != Y_NL) ;
}


/* Print register number N.  TYPE code indicate which register set to use.
   Return non-zero if register N was valid register string. */

static int
print_reg (reg_no, type_code)
     int reg_no;
     int type_code;
{
  switch (type_code)
    {
    case 0:
      write_output (message_out, "Reg %d = 0x%08x (%d)\n",
		 reg_no, R[reg_no], R[reg_no]);
      break;
    case 1:
      write_output (message_out, "FP reg %d = %f (double)\n",
		 reg_no, FPR_D (reg_no));
      write_output (message_out, "FP reg %d = %f (single)\n",
		 reg_no, FPR_S (reg_no));
      break;
    case 2:
      {
	char *s = (char *) reg_no, *s1 = (char *) reg_no;

	/* Conver to lower case */
	for ( ; *s != '\0'; s ++) *s = tolower (*s);
	/* Drop leading $ */
	if (*s1 == '$') s1 += 1;

	if (streq (s1, "pc"))
	  write_output (message_out, "PC = 0x%08x (%d)\n", PC, PC);
	else if (streq (s1, "hi"))
	  write_output (message_out, "HI = 0x%08x (%d)\n", HI, HI);
	else if (streq (s1, "lo"))
	  write_output (message_out, "LO = 0x%08x (%d)\n", LO, LO);
	else if (streq (s1, "fpcond"))
	  write_output (message_out, "FpCond = 0x%08x (%d)\n", FpCond, FpCond);
	else if (streq (s1, "cause"))
	  write_output (message_out, "Cause = 0x%08x (%d)\n", Cause, Cause);
	else if (streq (s1, "epc"))
	  write_output (message_out, "EPC = 0x%08x (%d)\n", EPC, EPC);
	else if (streq (s1, "status"))
	  write_output (message_out, "Status = 0x%08x (%d)\n",
		     Status_Reg, Status_Reg);
	else if (streq (s1, "badvaddr"))
	  write_output (message_out, "BadVAddr = 0x%08x (%d)\n",
		     BadVAddr, BadVAddr);
	else if (streq (s1, "context"))
	  write_output (message_out, "Context = 0x%08x (%d)\n",
			Context, Context);
	else if (streq (s1, "prid"))
	  write_output (message_out, "PRId = 0x%08x (%d)\n", PRId, PRId);
	else
	  return (0);
	break;
      }
    }
  return (1);
}



/* Print anx error message. */

/*VARARGS0*/
void
error (va_alist)		/* Display only */
va_dcl
{
  va_list args;
  char *fmt;

  va_start (args);
  fmt = va_arg (args, char *);
#if defined(mips) || defined(__linux__)
  vfprintf (stderr, fmt, args);
#else
  _doprnt (fmt, args, stderr);
#endif
  va_end (args);
}


/*VARARGS0*/
int
run_error (va_alist)		/* Display and return to top level */
va_dcl
{
  va_list args;
  char *fmt;

  va_start (args);
  fmt = va_arg (args, char *);
#if defined(mips) || defined(__linux__)
  vfprintf (stderr, fmt, args);
#else
  _doprnt (fmt, args, stderr);
#endif
  va_end (args);
  longjmp (spim_top_level_env, 1);
  return (0);			/* So it can be used in expressions */
}



/*VARARGS0*/
void
write_output (va_alist)
va_dcl
{
  va_list args;
  char *fmt;
  FILE *f;

  va_start (args);
  f = va_arg (args, FILE *);
  fmt = va_arg (args, char *);
#if defined(mips) || defined(__linux__)
  if (f != 0)
    vfprintf (f, fmt, args);
  else
    vfprintf (stdout, fmt, args);
#else
  _doprnt (fmt, args, f);
#endif
  fflush (f);
  va_end (args);
}


void
read_input (str, n)
     char *str;
     int n;
{
  fgets (str, n, stdin);
}
